/*
 * Created on Jan 8, 2005
 * 
 */
package com.cachingapp.caching;

import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import javax.servlet.ServletContextEvent;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheException;
import org.jboss.cache.PropertyConfigurator;
import org.jboss.cache.TreeCache;

import com.loanapp.LoanAppHelper;
import com.cachingapp.cache.CachingException;
import com.cachingapp.cache.ICacheKey;
import com.loanapp.listener.LoanAppContextListener;

/**
 * @author Administrator
 *  
 */
public aspect JBossCache extends ObjectCache {
    private static Log sLog = LogFactory.getLog(JBossCache.class);
    
    private TreeCache cache;
    private String props = null;
    private String cacheGroupName = "CacheGroup";
    private String fqnName = "/cacheaop/cache/object";

    pointcut contextInitialized(ServletContextEvent event) : args(event) 
    && (execution(void LoanAppContextListener.contextInitialized(ServletContextEvent)));

    pointcut contextDestroyed(ServletContextEvent event) : args(event) 
    && (execution(void LoanAppContextListener.contextDestroyed(ServletContextEvent)));

    pointcut getInterestRates(String productGroup) : args(productGroup)
    && (execution(List LoanAppHelper.getInterestRates(String)));
 
    public JBossCache() {
        //
    }
    
    public void initObjectCache() {
        try {
            cache = new TreeCache();
            PropertyConfigurator config = new PropertyConfigurator();
            config.configure(cache, "jbosscache-config.xml");
            cache.setClusterName(cacheGroupName);
            cache.createService();
            cache.startService();
            sLog.debug("cache : " + cache);
        } catch (Exception e) {
            // Handle cache region initialization failure
            sLog.debug("Error in creating JBossCache.");
            sLog.error(e);
        }
    }
    
    public void cleanupObjectCache() {
        cache.stopService();
    }
    
   
   // CACHE METHODS
    public List getInterestRates(String productGroup, String regionName) 
			throws CachingException {
        List interestRates = null;

        long t1 = System.currentTimeMillis();

        try {
            // First attempt to load from cache
            interestRates = (List)cache.get(fqnName, productGroup);
        } catch (Exception e) {
            sLog.error(e);
        }

        try {
            // The object was not found in the cache
            if (interestRates == null) {
                sLog.info("Could not find the object in cache for productGroup " + productGroup);
            } else {
                sLog.info("Found the object in cache for productGroup " + productGroup);
            }

            long t2 = System.currentTimeMillis();
            sLog.info("Time taken to get cached object is : " + (t2 - t1) + " milli-seconds.");

        } catch (Exception e) {
            throw new CachingException("Exception in getCacheObject()", e);
        }
        return interestRates;
    }
    
    /**
     * Stores CacheObjects in database. Clears old items and caches new.
     */
    public void putCacheObject(Object cacheKey, Object cacheObject,
            String regionName) throws CachingException {
        Object cacheObjectKey = null;
        try {
            // Check if cacheKey is an instance of CacheKey interface
            if (cacheKey instanceof ICacheKey) {
                cacheObjectKey = ((ICacheKey) cacheKey).getCacheKey();
            } else {
                cacheObjectKey = cacheKey;
            }
            // put the new object in the cache
            cache.put(fqnName, String.valueOf(cacheObjectKey), cacheObject);
        } catch (Exception e) {
            // Handle failure in removing object or putting object to cache.
            e.printStackTrace();
            sLog.error("Error in putting object into cache");
            throw new CachingException("Error in putting object into cache");
        }
    }

    /**
     * Peeks into the cache for a given key without loading the value on demand.
     * If the value is cached, then it will be returned. If the value is not
     * already cached, then null is returned.
     */
    public Object peek(Object cacheKey, String regionName)
            throws CachingException {
        Object cacheObject = null;
        Object cacheObjectKey = null;
        // Check if cacheKey is an instance of CacheKey interface
        if (cacheKey instanceof ICacheKey) {
            cacheObjectKey = ((ICacheKey) cacheKey).getCacheKey();
        } else {
            cacheObjectKey = cacheKey;
        }

        try {
            if (cache.get(fqnName, (String) cacheObjectKey) == null) {
                sLog.debug("Object does not exist in " + regionName + " region.");
            } else {
                sLog.debug("Object exists in the cache in " + regionName + " region.");
                cacheObject = cache.get(fqnName, String.valueOf(cacheObjectKey));
            }
        } catch (CacheException e) {
            sLog.error(e,e);
        }
        return cacheObject;
    }
    
    public Set getCacheObjectKeys(String fqnName) throws CachingException {
        sLog.debug("getCacheObjectKeys in JBossCache called..");
        Set keys = null;
        try {
	        keys = cache.getKeys(fqnName);
	        Iterator it = keys.iterator();
	        while (it.hasNext()) {
	            String key = it.next().toString();
	            String value = cache.get(fqnName, key).toString();
	            sLog.debug("cache key=" + key);
	            sLog.debug("cache value=" + value);
	        }
        } catch (Exception e) {
            sLog.error(e,e);
        }
        return keys;
    }    

    public void printCacheStatistics() throws CachingException {
        sLog.debug("printCacheStatistics in JBossCache called..");
        try {
	        sLog.debug("members=" + cache.getMembers().size());
	        Vector members = cache.getMembers();
	        for (int i=0; i<members.size(); i++) {
	            sLog.debug("member="+members.get(i).toString());
	        }
	        Set keys = cache.getKeys(fqnName);
	        Iterator it = keys.iterator();
	        while (it.hasNext()) {
	            String key = it.next().toString();
	            String value = cache.get(fqnName, key).toString();
	            sLog.debug("cache key=" + key);
	            sLog.debug("cache value=" + value);
	        }
        } catch (Exception e) {
            sLog.error(e,e);
        }
    }
}
