package com.loanapp.test;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class LoanAppClient extends Thread {

	private static Log sLog = LogFactory.getLog(LoanAppClient.class);

	private String host1;
	private int port1;
	private String host2;
	private int port2;
	private String webPage = "CachingAOP/LoanAppServlet";
	private int reps;
	private long delay;

	private String loanRequest = "mortgageRates";
	private java.util.Properties p = new java.util.Properties();

	private static long responsetime = 0;
	private static long numOfRequests = 0;
	private static long lastcheck = System.currentTimeMillis();
	private static long maxResponseTime = -1;
	private static long minResponseTime = Long.MAX_VALUE;
	private static int maxThreadCount = 0;
	private static int deadThreadCount = 0;

	private static int threadCounter = 0;
	private static int[] errorCounter = new int[1000];

	static {
		java.util.Arrays.fill(errorCounter, 0);
	}

	// CONSTRUCTOR
	public LoanAppClient(String name, String host1, int port1,
			String host2, int port2, int reps, long delay) {
		super(name);
		this.host1 = host1;
		this.port1 = port1;
		this.host2 = host2;
		this.port2 = port2;
		this.reps = reps;
		this.delay = delay;
	}

	public void run() {
		try {
			inc();
			runTest(host1, port1, host2, port2, reps, delay);
			LoanAppClient t = new LoanAppClient(this.getName(), host1, port1,
					host2, port2, reps, delay);
			t.setDaemon(false);
			t.start();
			sLog.debug("Thread " + this.getName() + " finished, started another one");
		} finally {
			dec();
		}
		printStats();
	}

	public void runTest(String host1, int port1, String host2, int port2, int reps, long delay) {
		java.util.Random r = new java.util.Random(System.currentTimeMillis());
		HttpClient httpClient = new HttpClient();

		int allOK = 1;
		int counter = 0;
		for (int i = 0; (i < reps) && (allOK >= 0); i++) {
			boolean doAdd = r.nextBoolean();
			if (p.size() > 20) {
				doAdd = false;
			}
			int index = (r.nextInt() % 8);
			if (index < 0) {
				index = (index * ( -1));
			}

			if (doAdd) {
				sLog.debug("runTest(): loanRequest=Mortgage");
				p.setProperty("loanRequest", "Mortgage");
				p.setProperty("points", "0");
				allOK = this.doRequest1(httpClient);
//				allOK = this.doRequest2(httpClient);
			} else {
			    p.clear();
				p.setProperty("loanRequest", "Heloc");
				p.setProperty("points", "1");
				allOK = this.doRequest1(httpClient);
			}
			counter++;
			try {
				//sLog.debug("Thread ["+Thread.currentThread().getName()+"] has received sessionId="+p.getProperty("sessionId"));
				Thread.sleep(delay);
			} catch (Exception ignore) {}
		} //for
		try {
			sLog.debug("Thread [" + Thread.currentThread().getName() +
					"] completed at index:" + counter);
			httpClient.endSession();
		} catch ( Exception x ) {
			x.printStackTrace();
		}
	}

	/**
	 *
	 * @param con
	 * @return 1 = all is ok
	 * @return 0 = try previous request
	 * @return -1 = failed
	 */
	public int doRequest1(HttpClient httpClient) {
		try {
			// Host1
			NameValuePair form_data[] = new NameValuePair[p.size()];
			java.util.Enumeration e = p.propertyNames();
			int index = 0;
			while (e.hasMoreElements()) {
				String name = (String) e.nextElement();
				String value = p.getProperty(name);
				form_data[index++] = new NameValuePair(name, value);
			} //while
			String strUrl = "http://" + host1 + ":" + port1 + "/" + webPage;
			sLog.debug("strUrl : " + strUrl);
			PostMethod method = new PostMethod(strUrl);
			method.setHttp11(false);
			method.setRequestBody(form_data);

			int statusCode = -1;
			int retries = 0;
			while (statusCode == -1) {
				try {
					sLog.debug("["+Thread.currentThread().getName()+"] Executing...");
					sLog.debug("["+Thread.currentThread().getName()+"] Request for SessionId="+p.getProperty("sessionId"));
					long start = System.currentTimeMillis();
					statusCode = httpClient.executeMethod(method);
					long stop = System.currentTimeMillis();
					addRspTime(stop - start, statusCode);
				} catch (Exception x) {
					try {
						Thread.currentThread().sleep(delay);
					} catch (Exception ignore) {}
					try {
						method.recycle();
						Thread.currentThread().sleep(10 * 1000);
					} catch (Exception ignore) {}
					if (retries > 10) {
						sLog.debug("[" + Thread.currentThread().getName() +
								"] Unable to execute command after 10 retries. Quitting:" +
						x.getMessage());
						x.printStackTrace();
						method.releaseConnection();
						return -1;
					}
				}
				retries++;
			}
			sLog.debug("statusCode : " + statusCode);
			return processStatusCode(statusCode, method);
		} catch (Exception ioe) {
			sLog.debug("[" + Thread.currentThread().getName() +
					"] Unable to receive data, lets try again:" +
			ioe.toString());
			return 0;
		}
	}

	public int doRequest2(HttpClient httpClient) {
		try {
			// Host2
			NameValuePair form_data[] = new NameValuePair[1];
			String name = "loanRequest";
			String value = "Mortgage";
			form_data[0] = new NameValuePair(name, value);

			String strUrl = "http://" + host2 + ":" + port2 + "/" + webPage;
			sLog.debug("strUrl : " + strUrl);
			PostMethod method = new PostMethod(strUrl);
			method.setHttp11(false);
			method.setRequestBody(form_data);

			int statusCode = -1;
			int retries = 0;
			while (statusCode == -1) {
				try {
					sLog.debug("["+Thread.currentThread().getName()+"] Executing...");
					sLog.debug("["+Thread.currentThread().getName()+"] Request for SessionId="+p.getProperty("sessionId"));
					long start = System.currentTimeMillis();
					statusCode = httpClient.executeMethod(method);
					long stop = System.currentTimeMillis();
					addRspTime(stop - start, statusCode);
				} catch (Exception x) {
					try {
						Thread.currentThread().sleep(delay);
					} catch (Exception ignore) {}
					try {
						method.recycle();
						Thread.currentThread().sleep(10 * 1000);
					} catch (Exception ignore) {}
					if (retries > 10) {
						sLog.debug("[" + Thread.currentThread().getName() +
								"] Unable to execute command after 10 retries. Quitting:" +
						x.getMessage());
						x.printStackTrace();
						method.releaseConnection();
						return -1;
					}
				}
				retries++;
			}
			sLog.debug("statusCode : " + statusCode);
			return processStatusCode(statusCode, method);

		} catch (Exception ioe) {
			sLog.debug("[" + Thread.currentThread().getName() +
					"] Unable to receive data, lets try again:" +
			ioe.toString());
			return 0;
		}
	}

	private int processStatusCode(int statusCode, PostMethod method) throws Exception {
		if (statusCode == 302) {
			sLog.debug("[" + Thread.currentThread().getName() + "] Received 302 Error, \n"
					+ printRespHeader(method.getResponseHeaders()) + " will resend same command again.");
			String s = method.getResponseBodyAsString();
			sLog.debug("ResponseBody :" + s);
			incError(302);
			method.releaseConnection();
			return 1;
		} else if (statusCode == 400) {
			sLog.debug("[" + Thread.currentThread().getName() + "] Received Error: " + method.getStatusLine() + " "
					+ method.getStatusText() + " " + method.getResponseBodyAsString()
					+ "\nWill try again...after 1 sec sleep");
			String s = method.getResponseBodyAsString();
			sLog.debug("ResponseBody :" + s);
			try {
				Thread.currentThread().sleep(1000);
			} catch (Exception x) {
			}
			incError(statusCode);
			method.releaseConnection();
			return 1;
		} else if (statusCode >= 300) {
			sLog.debug("[" + Thread.currentThread().getName() + "] Received Error: " + method.getStatusLine() + " "
					+ method.getStatusText() + " " + method.getResponseBodyAsString());
			String s = method.getResponseBodyAsString();
			sLog.debug("ResponseBody :" + s);
			incError(statusCode);
			method.releaseConnection();
			return -1;
		} else if (statusCode == 200) {
			String s = method.getResponseBodyAsString();
			sLog.debug("ResponseBody :" + s);
			java.io.StringBufferInputStream in = new java.io.StringBufferInputStream(s);
			java.util.Properties tmp = new java.util.Properties();
			tmp.load(in);
			if (tmp.getProperty("sessionId") != null) {
				p = tmp;
				incError(statusCode);
				method.releaseConnection();
				return 1;
			} else {
				incError(601);
				//sLog.debug("Received invalid (wrong page) response from
				// server="+statusCode+" trying again...");
				//if ( p.getProperty("sessionId")==null ) {
				//	sLog.debug("No session id in the properties");
				//	System.exit(0);
				//}
				method.releaseConnection();
				return 1;
			}
			//sLog.debug("["+Thread.currentThread().getName()+"] completed
			// request #"+p.getProperty("counter")+" Number of
			// values="+p.size()+"\nResponse="+s);
		} else {
			sLog.debug("[" + Thread.currentThread().getName() + "] Received unknown response=" + statusCode + ", \n"
					+ printRespHeader(method.getResponseHeaders()) + " will resend same command again.");
			method.releaseConnection();
			return -1;
		}
	}

	private synchronized static void addRspTime(long time, int statusCode) {
		if (statusCode == 200) {
			maxResponseTime = Math.max(time, maxResponseTime);
			minResponseTime = Math.min(time, minResponseTime);
		}
		responsetime += time;
		numOfRequests++;
		if ( (numOfRequests % 100) == 0) {
			printStats();
			lastcheck = System.currentTimeMillis();
		}
	}

	private static String getErrorCodes() {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < errorCounter.length; i++) {
			if (errorCounter[i] > 0) {
				buf.append("\n\tNr of ").append(i).append(" errors=").append(
				errorCounter[i]);
			}
		}
		return buf.toString();
	}

	private synchronized static void printStats() {
		sLog.debug("Thread count=" + getThreadCount() +
				"\t\t Max Thread count=" + maxThreadCount +
				"\t\t Average response time=" + getAverageResponseTime() +
				"\t\t for " + numOfRequests + " requests." +
				"\t\t Max response time(200 only)=" + maxResponseTime + "ms." +
				"\t\t Min response time(200 only)=" + minResponseTime + "ms." +
				getErrorCodes());

	}

	public synchronized static long getAverageResponseTime() {
		return responsetime / (numOfRequests == 0 ? 1 : numOfRequests);
	}

	public static synchronized void inc() {
		threadCounter++;
		maxThreadCount = Math.max(maxThreadCount, threadCounter);
		printStats();
	}

	public static synchronized void dec() {
		threadCounter--;
		deadThreadCount++;
		printStats();
	}

	public static synchronized void incError(int errorCode) {
		errorCounter[errorCode] = errorCounter[errorCode] + 1;
		if (errorCode != 200) {
			printStats();
		}
	}

	public static synchronized int getThreadCount() {
		return threadCounter;
	}

	public static String getPropertyString(java.util.Properties p) {
		StringBuffer buf = new StringBuffer();
		java.util.Enumeration e = p.propertyNames();
		while (e.hasMoreElements()) {
			String name = (String) e.nextElement();
			String value = p.getProperty(name);
			buf.append("\n\t").append(name).append("=").append(value);
		}
		return buf.toString();
	}

	public static String printRespHeader(Header[] headers) {
		StringBuffer buf = new StringBuffer("HEADERS:");
		for (int i = 0; i < headers.length; i++) {
			buf.append("\n\t");
			buf.append(headers[i].toExternalForm());
		}
		return buf.toString();
	}

	public static void main(String[] args) throws Exception {
		sLog.debug("Usage org.apache.catalina.cluster.test.client.LoanAppClient\n\t" +
				"numberofthreads host1 port1 host2 port2 reps delay\n\t" + "Example:\n\t" +
				"java org.apache.catalina.cluster.test.client.LoanAppClient 3 localhost 9080 localhost 10080 1000 100");

		int numOfThreads = Integer.parseInt(args[0]);
		sLog.debug("numOfThreads : " + numOfThreads);
		String host1 = args[1];
		sLog.debug("Host1 : " + host1);
		int port1 = Integer.parseInt(args[2]);
		sLog.debug("port1 : " + port1);
		String host2 = args[3];
		sLog.debug("Host2 : " + host2);
		int port2 = Integer.parseInt(args[4]);
		sLog.debug("port2 : " + port2);
		int reps = Integer.parseInt(args[5]);
		sLog.debug("reps : " + reps);
		long delay = Long.parseLong(args[6]);
		sLog.debug("delay : " + delay);
		sLog.debug("Cluster test started at " + new java.util.Date());
		for (int i = 0; i < numOfThreads; i++) {
			LoanAppClient t = new LoanAppClient("LoanAppClient #" + i, host1,
			port1, host2, port2, reps, delay);
			t.setDaemon(false);

			t.start();
			Thread.currentThread().sleep(delay / 3);
		}
	}

}
